/*******************************************************************
**                                                                **
**                       cell_sheet.c                             **
**                Written by U.S.Bhalla.                          **
**  Permission to use and modify this software is freely granted, **
**    provided this message and acknowledgement remain intact.    **
**                                                                **
**  The cell-creation code is derived from code written by        **
**                      Matt Wilson                               **
**                                                                **
**  I do not take any responsibility for the functionality and    **
**  use of this code, which is provided 'as is'.                  **
**                                                                **
*******************************************************************/

/*
** Cell_sheet : creates a sheet of cells based on coords specified
** by a coordinate file as a list of numbers. The cells are copies
** of prototype cells. Options allow one to reorient the cells to
** remain normal to the local plane of cells, to reposition the
** compartments of the cells to follow the curve of the plane,
** and to randomise the coordinates as the cells are made.
*/


#include <stdio.h>
#include <math.h>
#include "sim_ext.h"
#include "tools.h"

#define N_SURFACE_PTS 5000 /* A really big simulation */
#ifndef RAXIAL
#define RAXIAL 1
#endif

#ifndef dot_prd
#define dot_prd(x1,y1,z1,x2,y2,z2) (x1 * x2 + y1 * y2 + z1 * z2)
#endif

typedef struct surface_pt_type {
	float x,y,z;	/* coords of pt */
	float nx,ny,nz;	/* vector for normal */
} SurfacePt;

do_cell_sheet(argc,argv)
int argc;
char **argv;
{
Element 	*element;
Element 	*new_element;
Element		*last_element;
Element 	*parent;
Element 	*new_parent;
Element		*elm;
ElementList	sel,el;
Element		*selms[SMALL_EL_SIZE],*elms[MAX_EL_SIZE];
FILE		*fp,*fopen();
int			i,j;
int			orient = 0, curve = 0, randomise = 0,twirl = 0;
char		*str;
float		fx,fy,fz;
char		name[100];
int			index;
char		*ptr;
SurfacePt	pts[N_SURFACE_PTS];
SurfacePt	*pt;
int			npts;
float		theta;

	sel.nelements = 0;
	sel.element = selms;
	el.nelements = 0;
	el.element = elms;

    if(argc < 4){
	printf("usage: %s prototype dest filename [o(rient)][c(urve)][r(andomise) x y z]\n",
	argv[0]);
	printf("[t(wirl) theta\n");
	printf("The 'prototype' cell is copied to the 'dest' element to\n");
	printf("form a sheet of cells at positions specified in three-d\n");
	printf("by the file. The 'orient' option reorients the cell to\n");
	printf("be normal to the local region of the sheet, The normals are also in the file\n");
	printf("The 'curve' option positions the children elements\n");
	printf("of the cell to follow the curvature of the sheet\n");
	printf("The randomise option randomises the coordinates by the\n");
	printf("specified factors in x,y,z before orienting/positioning\n");
	printf("The twirl option rotates each cell about the normal by\n");
	printf("a random angle from 0 to theta (in degrees)\n");
	return;
    }
	for (i = 4; i < argc ; i++) {
		str = argv[i];
		if (str[0] == '-')
			str++;
		if(*str == 'o') /* orient */
			orient = 1;
		if(*str == 'c') /* curve */
			curve = 1;
		if(*str == 'r') /* randomise */ {
			if (argc < (i + 3)) {
				printf("usage: %s prototype dest filename [o(rient)][c(urve)][r(andomise) x y z]\n",argv[0]);
				return;
			}
			randomise = 1;
			i++; fx = atof(argv[i]);
			i++; fy = atof(argv[i]);
			i++; fz = atof(argv[i]);
		}
		if(*str == 't') /* twirl */ {
			if (argc < (i + 1)) {
				printf("usage: t(wirl) theta\n");
				return;
			}
			twirl = 1;
			i++; theta = atof(argv[i]);
			theta *= PI/180.0;
		}
	}

    /*
    ** open the sheet file
    */
	if (!(fp = fopen(argv[3],"r"))) {
		fprintf(stderr,"could not open coordinate file '%s'\n",argv[3]);
		return;
	}
	printf("Reading in cell coordinates from file %s\n",argv[3]);
    /*
    ** find the element used to fill the map
    */
    element = GetElement(argv[1]);
    /*
    ** find the element to attach the elements to
    */
    parent = GetElement(argv[2]);
    /*
    ** if no such parent exists
    */
    if(parent == NULL){
	/*
	** try to create it using the next level up on the tree as
	** the new parent for the parent
	*/
		strcpy(name,GetParentComponent(argv[2]));
		new_parent = GetElement(name);
		/*
		** if some place to attach the parent was found then go ahead
		** and create the parent
		*/
		if(new_parent != NULL){
			strcpy(name,GetBaseComponent(argv[2]));
			index = GetTreeCount(name);
			if (ptr = strchr(name,'[')) {
				*ptr = '\0';
			} 
	   		parent = Create("neutral",name,new_parent,NULL,index);
		} else {
			printf("could not create element %s\n",name);
			return;
		}
    }
	last_element = parent->child;

	for(npts = 0, pt = pts ;
		fscanf(fp,"%f%f%f%f%f%f",
		&(pt->x),&(pt->y),&(pt->z),&(pt->nx),&(pt->ny),&(pt->nz)) == 6;
		npts++, pt++)
		;
	fclose(fp);

	for (i = 0 , pt = pts; i < npts ; pt++,i++) {
	    new_element = CopyElementTree(element);
	    AttachToEnd(parent,last_element,new_element);
		if (last_element == NULL) {
			new_element->index = 0;
		} else {
			new_element->index = last_element->index + 1;
		}
		last_element = new_element;
		CopyConnections(element,new_element);
		CopyMsgs(element,new_element);

		if (randomise)
			rancoord(new_element,fx,fy,fz,1,1);

		if (twirl) {
			find_somas(new_element,&sel);
			for(j = 0 ; j < sel.nelements ; j++) {
				elm = sel.element[j];
				find_all_dends(elm,1,&el);
				RotateAboutAxis(&el, elm->x,elm->y,elm->z, 0.0,0.0,1.0,
					frandom(theta,-theta));
			}
		}

		rel_position(new_element,pt->x,pt->y,pt->z);

		if (orient) {
			find_all_kids(new_element,&el);
			rotate3d(&el,pt->x,pt->y,pt->z,
				0.0,0.0,1.0,
				pt->nx,pt->ny,pt->nz);
		}

		if (curve) {
			curve_somas(new_element,pt,pts,npts);
		}
	}
	free(pts);
	npts = 0;
	printf("coordinate file %s read\n",argv[3]);
}

/*
** If elm is not a compartment, curve looks for somas among its children
** to do the curving algorithm from.
*/
curve_somas(elm,elmpt,pts,npts) 
	Element	*elm;
	SurfacePt	*elmpt,*pts;
{
    Element *child;
	ElementList	el;
	Element	*elms[MAX_EL_SIZE];
	int		i;
	
	el.nelements = 0;
	el.element = elms;
	find_somas(elm,&el);
	
	for(i = 0 ; i < el.nelements ; i++)
		curve_arbor(el.element[i],elmpt,pts,npts);
}

/*
** curve_arbor : adjusts the arbor to conform to the curvature of the
** surface  : for now, assume it is only convex.
** This algorithm assumes that the length of each compartment is less
** than the radius of curvature of the surface, otherwise the errors
** tend to add up. It also assumes that the sampling density over
** the plane results in a smaller spacing of samples than the 
** dendritic length.
** This routine does not change the positions of compartments which
** are within a cylinder of xxx from the parent elm.
*/
curve_arbor(elm,elmpt,pts,npts)
	Element	*elm;
	SurfacePt	*elmpt,*pts;
	int		npts;
{
	SurfacePt	pt;
	ElementList	el;
	Element		*elms[MAX_EL_SIZE];
	MsgIn		*msg;
	Element		*kid;
	float		theta;
	float		c;

	el.element = elms;
	el.nelements = 0;
/* Find ref coord, where normal to elm crosses surface. */
/* Move out from elm to dends, ignoring all compartments less than
** pi/6 from normal. */
/* Find closest pt to dend. */
/* Get angle between line joining this pt and elm, and normal. */
/* Rotate elm and all kids by this angle, in plane defined by 
** normal, parent elm, and new elm. */
/* Continue recursively for all kids down this branch, now defining
** parent elm as the rotated elm, and the normal as the perp to the
** dend in the plane of the original normal. */

	/* Scan through all kids of element */
	for (msg=elm->msg_in;msg;msg = msg->next) {
		if (msg->type == RAXIAL) {
			kid = msg->src;
			/* find the nearest pt in surface definition list */
			find_nearest_pt(pts,npts,kid->x,kid->y,kid->z,&pt);

			/* Find the angle between the normal to this point and
			** the original normal */
			c = dot_prd(pt.nx,pt.ny,pt.nz,
				elmpt->nx,elmpt->ny,elmpt->nz);
			theta = acos(c);
			/* If theta is not between 10 and 60 degrees, skip curve */
			if (theta < (PI/18.0) || theta > (PI/3.0)) {
				curve_arbor(kid,&pt,pts,npts);
				continue;
			}

			/* Sign flip since we wish to rotate DOWN, not up */
			theta = -theta;

			/* Rotate 'kid' and all its subkids by theta */
			find_all_dends(kid,1,&el);
			RotateInPlane(&el,elm->x,elm->y,elm->z,kid->x,kid->y,kid->z,
				elm->x + elmpt->nx, elm->y + elmpt->ny,
				elm->z + elmpt->nz,theta);
			/* Recurse down to repeat whole process for kid */
			curve_arbor(kid,&pt,pts,npts);
		}
	}
}

find_nearest_pt(pts,npts,x,y,z,pt)
	SurfacePt	*pts;
	int			npts;
	float		x,y,z;
	SurfacePt	*pt;
{
	register int		i;
	register float	r;
	int	besti = -1;
	register float	bestr;
	register	SurfacePt	*spt;

	spt = pts;
	bestr = (spt->x - x) * (spt->x - x) +
			(spt->y - y) * (spt->y - y) +
			(spt->z - z) * (spt->z - z);

	for (i = 0, spt = pts ; i < npts ; i++, spt++) {
		r = (spt->x - x) * (spt->x - x) +
			(spt->y - y) * (spt->y - y) +
			(spt->z - z) * (spt->z - z);
		if (r < bestr) {
			bestr = r;
			besti = i;
		}
	}
	if (besti < 0)
		*pt = pts[0];
	else 
		*pt = pts[besti];
}
